/* Implement basic draw primitives: pixels, lines and frame buffer operations */



#include <string.h>
#include <math.h>

#include "../defs.h"

#include "../mem.h"

#include "../string.h"

#include "../texture.h"

#include "../lighttab.h"

#include "../asm/draw.inc"

#include "draw.h"



/* Set all frame buffer with a specific color.
**********************************************************/
void CDraw::mClear ( BYTE yCol )
{
   memset ( (void*) *m_poFrBuff, yCol, m_nSize ); // (void*) *m_poFrBuff is
                                         // one way to access memory buffer.
}



/* Put one pixel on frame buffer.
**********************************************************/
void CDraw::mSetPixel ( INT nX, INT nY, BYTE yCol )
{
   // do clip
   if ( nX >= 0 && nX < m_nResX && nY >= 0 && nY < m_nResY )
      memset ( (void*) ( m_poFrBuff->m_payMem + nX + ( nY * m_nResX ) ), yCol, 1 );
}



/* Copy an other frame buffer over the current frame buffer.
oFrBuffSource is the source frame buffer and must have
the same size as the current frame buffer.
**********************************************************/
void CDraw::mCopy ( CMem& oFrBuffSource )
{
   memcpy ( (void*) *m_poFrBuff, (void*) oFrBuffSource, m_poFrBuff->m_dwSize );
}



/* 2 - Copy an other frame buffer (or bitmap) over the
current frame buffer.
(nOX, nOY) - destin frame buffer position where place
the upper left source bitmap point (these values start
in upper left frame buffer corner).
nWidth and nHeidht - the dimensions of the bitmap.
No test if bitmap is bigger that the frame buffer in any axis (GPF)!
**********************************************************/
void CDraw::mCopy ( INT nOX, INT nOY, INT nWidth, INT nHeight, CMem& oFrBuffSource )
{
   // copy each bitmap line, no clip!
   for ( INT ncY = 0; ncY < nHeight; ncY ++ )

      memcpy ( (void*) ( m_poFrBuff->m_payMem + ( ncY * m_nResX ) + ( nOY * m_nResX ) + nOX ),
               (void*) ( oFrBuffSource.m_payMem + ncY * nWidth ),
               nWidth );
}



/* Draw a line with the specific color. Low optimized code.
Given points (nXI, nYI) -> (nXF,nYF). Do line clip (pixel by pixel).
**********************************************************/
void CDraw::mLine ( INT nXI, INT nYI, INT nXF, INT nYF, BYTE yCol )
{
   // define start and end Y (the less and greater Y)
   int
      nYStart = nYI,
      nYEnd = nYF,
      nSignalY = 1;
   if ( nYF < nYStart ) {
      nYStart = nYF;
      nYEnd = nYI;
      nSignalY = -1;
   }

   // define start and end X (the less and greater X)
  int
      nXStart = nXI,
      nXEnd = nXF,
      nSignalX = 1;
   if ( nXF < nXStart ) {
      nXStart = nXF;
      nXEnd = nXI;
      nSignalX = -1;
   }

   // compute delta between the given points
  int
    nDeltaX = nXF - nXI,
    nDeltaY = nYF - nYI;

   // if delta X is the greater you draw the line using Y as function of X
   if ( abs ( nDeltaX ) > abs ( nDeltaY ) ) {

      if ( nDeltaX == 0 )
         return; // BUGFIX

      float
        fYRate = (float) nDeltaY / nDeltaX, // define Y rate
          fY;                                            // start Y
      // define start line
      if ( nSignalY == -1 )
          fY = (float) nYEnd;
     else
       fY = (float) nYStart;

     // draw line
      // test if X is growing left (1) or right (-1)
      if ( nSignalX == 1 )
         for ( int nX = nXStart; nX < nXEnd + 1; nX ++ ) {

            memset ( (void*) ( m_poFrBuff->m_payMem + nX + ( (INT) fY * m_nResX ) ), yCol, 1 );

            fY += fYRate;
         }
      else {

        fYRate = - fYRate;

         for ( int nX = nXEnd; nX > nXStart; nX -- ) {

            memset ( (void*) ( m_poFrBuff->m_payMem + nX + ( (INT) fY * m_nResX ) ), yCol, 1 );

            fY += fYRate;
         }
      }
   }
   else {   // draw the line using X as function of Y

      if ( nDeltaY == 0 )
         return; // BUGFIX

      float
        fXRate = (float) nDeltaX / nDeltaY, // define X rate
          fX;                                            // start X
      // define start colum
      if ( nSignalX == -1 )
          fX = (float) nXEnd;
      else
          fX = (float) nXStart;

     // draw line
      // test if Y is growing down (1) or up (-1)
      if ( nSignalY == 1 )
         for ( int nY = nYStart; nY < nYEnd + 1; nY ++ ) {

            memset ( (void*) ( m_poFrBuff->m_payMem + (INT) fX + ( nY * m_nResX ) ), yCol, 1 );

            fX += fXRate;
         }
      else {

        fXRate = - fXRate;

         for ( int nY = nYEnd; nY > nYStart; nY -- ) {

            memset ( (void*) ( m_poFrBuff->m_payMem + (INT) fX + ( nY * m_nResX ) ), yCol, 1 );
 
            fX += fXRate;
         }
      }
   }
}



/* Set frame buffer data
return TRUE - ok \ FALSE - error
**********************************************************/
BOOL CDraw::mSetFrameBuffer ( INT nResX, INT nResY, CMem& oFrBuff )
{
   m_nResX = nResX;
   m_nResY = nResY;
   m_poFrBuff = &oFrBuff;
   m_nSize = m_nResX * m_nResY;
   m_nMaxX = m_nResX - 1;
   m_nMaxY = m_nResY - 1;
   m_nMidX = m_nResX >> 1;  //    >> 1 is equal / 2
   m_nMidY = m_nResY >> 1;

   // verify if need allocate memory
   if ( m_nResY > m_nNumLinesAlloc ) {
   
      // desalocate
      delete m_pastForwardEdges;
      delete m_pastBackwardEdges;

      // try alocate memory for edge buffer for forward scan conversion
      if ( ( m_pastForwardEdges = new tagPolyEdge [ m_nResY ] ) == NULL ) {
         printf ( MSG_ERROR_MEMORY_LOW );
         printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
         return FALSE;
      }
 
       // try alocate memory for edge buffer for backward scan conversion
      if ( ( m_pastBackwardEdges = new tagPolyEdge [ m_nResY ] ) == NULL ) {
         printf ( MSG_ERROR_MEMORY_LOW );
         printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
         return FALSE;
      }
      
      m_nNumLinesAlloc = m_nResY;
   }

   return TRUE;
}

