/* Affine textured polys with any number of vertexes and flat shading. */



#include <string.h>
#include <math.h>

#include "../defs.h"

#include "../mem.h"

#include "../string.h"

#include "../texture.h"

#include "../lighttab.h"

#include "../asm/draw.inc"

#include "draw.h"



/* Draw a affine texture mapped polygon on frame buffer.
The poly is flat shaded using the lighttable pointed by m_poLightTable.
Make clip through the given edge buffer (portal).
pastPolVertexes is the array of polygon vertexes;
pastTexVertexes is the array of texture vertexes;
One vertex in polygon has one vertex in texture.
All vertexes must be in clockwise or anticlockwise order.
nNumVertex is the number of polygon vertexes.
oTx is the texture to use in polygon.
fNormLightLevel has the normalized polygon light level (between 0 and 1)
pastEdgeBuffer is the clip buffer to be used (in portal clip),
 if set to NULL only frame buffer clip is made (default NULL).
**********************************************************/
void CDraw::mTexturedPolygon ( tagXY* pastPolVertexes,
                               tagXY* pastTexVertexes,
                               INT nNumVertexes, CTexture& oTx,
                               tagInterval* pastEdgeBuffer,
                               float fNormLightLevel )
{
   // verify if have enough number of vertexes
   if ( nNumVertexes < 3 )
      ABORT ( MSG_ERROR_LOW_VERTEXES, __FILE__, __LINE__ )

   // seach top and bottom vertexes
   INT
      nYMin = pastPolVertexes [ 0 ].nY,   // y of the top vertex
      nYMax = pastPolVertexes [ 0 ].nY,   // y of the bottom vertex
      nPosMin = 0,       // the index of the top vertex (upper in screen)
      nPosMax = 0;       // the index of the bottom vertex (lower in screen)
   // verify vertexes
   for ( INT ncVert = 1; ncVert < nNumVertexes; ncVert ++ ) {

      // verify if array vertex is greater than the current bottom vertex
      if ( pastPolVertexes [ ncVert ].nY > nYMax ) {
         nYMax = pastPolVertexes [ ncVert ].nY;
         nPosMax = ncVert;
      }

      // verify if array vertex is less than the current top vertex
      if ( pastPolVertexes [ ncVert ].nY < nYMin ) {
         nYMin = pastPolVertexes [ ncVert ].nY;
         nPosMin = ncVert;
      }
   }

   // verify if top and bottom are the same, if yes abort polygon fill
   if ( nYMax == nYMin )
      return;

   // scan forward in array of vertexes starting on top and going down
   // With this you scan one side of the polygon.
   INT
      nPos1 = nPosMin,         // the first vertex to be tested
      nPos2 = nPos1;           // the next vertex (used in same test)
   while ( nPos1 != nPosMax ) {

      // define the next vertex position in vertex array
      nPos2 ++;
      // if no more vertexes go to array start
      if ( nPos2 == nNumVertexes )
         nPos2 = 0;

      // define polygon and texture edges, the X for each scanline defined
      // by the line vertex 1 ----> vertex 2
      mScanTexturedEdge ( m_pastForwardEdges,
                          pastPolVertexes [ nPos1 ],
                          pastPolVertexes [ nPos2 ],
                          pastTexVertexes [ nPos1 ],
                          pastTexVertexes [ nPos2 ] );

      // define current vertex as the next vertex
      nPos1 = nPos2;
   }

   // scan backward in array of vertexes starting on top and going down
   // With this you scan one side of the polygon.
   nPos1 = nPosMin;         // the first vertex to be tested
   nPos2 = nPos1;           // the next vertex (used in same test)
   while ( nPos1 != nPosMax ) {

      // define the previous vertex position in vertex array
      nPos2 --;
      // if no more vertexes go to array end
      if ( nPos2 == -1 )
         nPos2 = nNumVertexes - 1;

      // define polygon and texture edges, the X for each scanline defined
      // by the line vertex 1 ----> vertex 2
      mScanTexturedEdge ( m_pastBackwardEdges,
                          pastPolVertexes [ nPos1 ],
                          pastPolVertexes [ nPos2 ],
                          pastTexVertexes [ nPos1 ],
                          pastTexVertexes [ nPos2 ] );

      // define current vertex as the previous vertex
      nPos1 = nPos2;
   }

   // define the light level for the loaded light table
   // transform normalized light level to light table value
   DWORD
      dwLightLevel = (DWORD) ( fNormLightLevel * ( m_poLightTable->m_dwNumLevels - 1 ) );

   // draw polygon
   // define scanline Y (if nYMin is less than 0, outside frame buff, use 0)
   INT
      nY = ( nYMin < 0 ) ? 0 : nYMin;
   // define memory position
   PTRVAL
      pLineAddress = (PTRVAL) m_poFrBuff->m_payMem + nY * m_nResX;

   // verify if clipping edge buffer exist, if yes use else use default
   tagInterval
      stDefaultClip;
   BYTE
      yNextLine = 1; // to go to next clip buffer element when scanning lines
   // if don't has a clip buffer (no portal clipping)
   if ( pastEdgeBuffer == NULL ) {
      // define the default clip
      stDefaultClip.nStart = 0;
      stDefaultClip.nEnd = m_nMaxX;
      pastEdgeBuffer = &stDefaultClip;
      // no line change
      yNextLine = 0;
   }
   else
      // find the right position to start buffer scan
      pastEdgeBuffer += nY;

   // if X values for this scanline are swapped
   if ( m_pastBackwardEdges [ nY ].nX < m_pastForwardEdges [ nY ].nX )
      // while Y is upper than bottom vertex or bottom frame buffer
      while ( nY < nYMax && nY < m_nResY ) {

         mHorizontalTexturedLine ( m_pastBackwardEdges [ nY ],
                                   m_pastForwardEdges [ nY ],
                                   *pastEdgeBuffer,
                                   (BYTE*) oTx, pLineAddress,
                                   dwLightLevel );
         // goto next scanline
         nY ++;
         pLineAddress += m_nResX;
         pastEdgeBuffer += yNextLine;
      }
   else
      // while Y is upper than bottom vertex or bottom frame buffer
      while ( nY < nYMax && nY < m_nResY ) {

         mHorizontalTexturedLine ( m_pastForwardEdges [ nY ],
                                   m_pastBackwardEdges [ nY ],
                                   *pastEdgeBuffer,
                                   (BYTE*) oTx, pLineAddress,
                                   dwLightLevel );

         // goto next scanline
         nY ++;
         pLineAddress += m_nResX;
         pastEdgeBuffer += yNextLine;
      }
}
   


/* Used by CDraw::mTexturedPolygon to define the X coordenates
in each scanline for the polygon and texture edges.
pastEdge - the x coordenates for each scanline in polygon and
           the x and y coordenates for each scanline in texture;
stP1 and stP2 - polygon vertex between scan must be done.
stT1 and stT2 - texture vertex between scan must be done.
Always stP1 is less (topper) than stP2 in Y axis.
Use fix point 16.16.
**********************************************************/
void CDraw::mScanTexturedEdge ( tagPolyEdge* pastEdge,
                                tagXY& stP1, tagXY& stP2,
                                tagXY& stT1, tagXY& stT2 )
{
   // verify if Y are in same line, if yes exit
   if ( stP1.nY == stP2.nY )
      return;

   INT
      // define where X start in first scanline
      nX = stP1.nX << 16,
      // define where texture X and Y start in first scanline
      nTexX = stT1.nX << 16,
      nTexY = stT1.nY << 16,
      // define X increase rate for each scanline
      nXInc = ( ( stP2.nX - stP1.nX ) << 16 ) / ( stP2.nY - stP1.nY ),
      // define texture X and Y increase rates for each scan line
      nTexXInc = ( ( stT2.nX - stT1.nX ) << 16 ) / ( stP2.nY - stP1.nY ),
      nTexYInc = ( ( stT2.nY - stT1.nY ) << 16 ) / ( stP2.nY - stP1.nY );

   // do Y clip in frame buffer limits
   INT
      nYTop,
      nYBottom;
   // if y for vertex 1 is upper than frame buffer limits
   if ( stP1.nY < 0 ) {
      nYTop = 0;
      // top Y has been cliped then is necessary update values
      nX    -= stP1.nY * nXInc;
      nTexX -= stP1.nY * nTexXInc;
      nTexY -= stP1.nY * nTexYInc;
   }
   else
      nYTop = stP1.nY;
   // if y for vertex 2 is lower than frame buffer limits
   if ( stP2.nY > m_nMaxY )
      nYBottom = m_nMaxY;
   else
      nYBottom = stP2.nY;

   // verify if need clip
   if ( nYBottom <= nYTop )
      return;

   // define X values in each scanline between top vertex and bottom vertex
   INT
      nYScan = nYTop;    // scanline in test

   do {

      pastEdge [ nYScan ].nX = nX >> 16;  // X in frame buff, convert fix->int
      pastEdge [ nYScan ].stTex.nX = nTexX; // X in texture
      pastEdge [ nYScan ].stTex.nY = nTexY; // Y in texture
      nX += nXInc;                       // new X in next scanline
      nTexX += nTexXInc;                 // new texture X
      nTexY += nTexYInc;                 // new texture Y
      nYScan ++;                         // next scanline

   } while ( nYScan <= nYBottom );       // while not found bottom vertex
}



/* Draw a horizontal textured line.
Clip X values through given clip edge X limits.
stEdgeIni - left edge values (X and texture start)
stEdgeEnd - right edge values (X and texture end)
stLimits - the X limits to clip the line
pyTex - texture; pLineAddress - where draw line in frame buffer.
pLineAddress is the position in frame buffer where the
dwLightLevel - the light level in current light table.
line must be drawn.
**********************************************************/
void CDraw::mHorizontalTexturedLine ( tagPolyEdge& stEdgeIni,
                                      tagPolyEdge& stEdgeEnd,
                                      tagInterval& stLimits,
                                      BYTE* pyTex, PTRVAL pLineAddress,
                                      DWORD dwLightLevel )
{
   // clip x values
   INT
      nFrXIni = stEdgeIni.nX,   // define left X value in frame buffer
      nFrXEnd = stEdgeEnd.nX;   // define right X value in frame buffer
   // verify start
   if ( stEdgeIni.nX < stLimits.nStart )
      nFrXIni = stLimits.nStart;
   // verify end
   if ( stEdgeEnd.nX > stLimits.nEnd )
      nFrXEnd = stLimits.nEnd;

   // verify if line has no lenght
   if ( nFrXEnd <= nFrXIni )
      return;

   INT
      // define where texture X and Y start
      nTexX = stEdgeIni.stTex.nX,
      nTexY = stEdgeIni.stTex.nY,
      // texture X and Y increase rates
      nTexXInc = ( stEdgeEnd.stTex.nX - stEdgeIni.stTex.nX ) / ( stEdgeEnd.nX - stEdgeIni.nX ),
      nTexYInc = ( stEdgeEnd.stTex.nY - stEdgeIni.stTex.nY ) / ( stEdgeEnd.nX - stEdgeIni.nX );

   // if X has been clipped
   if ( stEdgeIni.nX < stLimits.nStart ) {
      nTexX -= ( stEdgeIni.nX - stLimits.nStart ) * nTexXInc;
      nTexY -= ( stEdgeIni.nX - stLimits.nStart ) * nTexYInc;
   }

   // define start and end address in frame buffer
   BYTE
      *pyCurr = (BYTE*) pLineAddress + nFrXIni,   // current address
      *pyEnd = (BYTE*) pLineAddress + nFrXEnd;    // end address

   // line draw loop
   do {

      // compile making lighting
      #ifdef COMP_LIGHTS
      // light it

      // C only
      *pyCurr = * ( m_poLightTable->m_payLightTable +

                    // ( nTexY >> 8 ) & 0xFF00 == MAX_COLORS * ( nTexY >> 16 )
                    * ( pyTex + ( ( nTexY >> 8 ) & 0xFF00 ) + ( nTexX >> 16 ) ) +
                   
                    ( dwLightLevel << MAX_COLORS_EXP ) ); // * 256  ==  << 8


      #else    // no lights


      // compile with asm code
      #ifdef COMP_ASM
      ASM_TRANSFER_TEXEL ( pyCurr, pyTex, nTexX, nTexY );
      #else
      // C only
      *pyCurr = * ( pyTex + ( ( nTexY >> 8 ) & 0xFF00 ) + ( nTexX >> 16 ) );
      #endif

      #endif

      // update texture coordenates
      nTexX += nTexXInc;
      nTexY += nTexYInc;
      // go to next pixel
      pyCurr ++;

   } while ( pyCurr <= pyEnd );  // while not found end of line
}

